@@ -1,9 +1,10 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 from itertools import groupby, chain
 from functools import partial
-from sql import Table
+
+from sql import Table, Null
 from sql.functions import Overlay, Position
 from sql.operators import Concat
 
@@ -16,6 +17,9 @@ from trytond.pyson import If, Eval, Bool, PYSONEncoder, Id
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
+from trytond.modules.account.tax import TaxableMixin
+from trytond.modules.product import price_digits
+
 __all__ = ['Sale', 'SaleIgnoredInvoice', 'SaleRecreatedInvoice',
     'SaleLine', 'SaleLineTax', 'SaleLineIgnoredMove',
     'SaleLineRecreatedMove', 'SaleReport', 'OpenCustomer',
@@ -27,7 +31,7 @@ __metaclass__ = PoolMeta
 _ZERO = Decimal(0)
 
 
-class Sale(Workflow, ModelSQL, ModelView):
+class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     'Sale'
     __name__ = 'sale.sale'
     _rec_name = 'reference'
@@ -91,8 +95,8 @@ class Sale(Workflow, ModelSQL, ModelView):
         depends=['state'])
     currency = fields.Many2One('currency.currency', 'Currency', required=True,
         states={
-            'readonly': (Eval('state') != 'draft') |
-                (Eval('lines', [0]) & Eval('currency', 0)),
+            'readonly': ((Eval('state') != 'draft')
+                | (Eval('lines', [0]) & Eval('currency', 0))),
             },
         depends=['state'])
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
@@ -167,12 +171,19 @@ class Sale(Workflow, ModelSQL, ModelView):
         'get_shipment_returns', searcher='search_shipment_returns')
     moves = fields.Function(fields.One2Many('stock.move', None, 'Moves'),
         'get_moves')
+    origin = fields.Reference('Origin', selection='get_origin', select=True,
+        states={
+            'readonly': Eval('state') != 'draft',
+            },
+        depends=['state'])
 
     @classmethod
     def __setup__(cls):
         super(Sale, cls).__setup__()
-        cls._order.insert(0, ('sale_date', 'DESC'))
-        cls._order.insert(1, ('id', 'DESC'))
+        cls._order = [
+            ('sale_date', 'DESC'),
+            ('id', 'DESC'),
+            ]
         cls._error_messages.update({
                 'invalid_method': ('Invalid combination of shipment and '
                     'invoicing methods on sale "%s".'),
@@ -305,8 +316,8 @@ class Sale(Workflow, ModelSQL, ModelView):
                                 sale_line.id))
                         ).select(sql_table.id,
                             where=(sql_table.state == 'confirmed')
-                            & ((sale_line_invoice_line.id != None)
-                                | (move.id != None)))
+                            & ((sale_line_invoice_line.id != Null)
+                                | (move.id != Null)))
             cursor.execute(*sql_table.update(
                     columns=[sql_table.state],
                     values=['processing'],
@@ -375,32 +386,14 @@ class Sale(Workflow, ModelSQL, ModelView):
 
     @fields.depends('party', 'payment_term')
     def on_change_party(self):
-        invoice_address = None
-        shipment_address = None
-        payment_term = None
+        self.invoice_address = None
+        self.shipment_address = None
+        self.payment_term = self.default_payment_term()
         if self.party:
-            invoice_address = self.party.address_get(type='invoice')
-            shipment_address = self.party.address_get(type='delivery')
+            self.invoice_address = self.party.address_get(type='invoice')
+            self.shipment_address = self.party.address_get(type='delivery')
             if self.party.customer_payment_term:
-                payment_term = self.party.customer_payment_term
-
-        changes = {}
-        if invoice_address:
-            changes['invoice_address'] = invoice_address.id
-            changes['invoice_address.rec_name'] = invoice_address.rec_name
-        else:
-            changes['invoice_address'] = None
-        if shipment_address:
-            changes['shipment_address'] = shipment_address.id
-            changes['shipment_address.rec_name'] = shipment_address.rec_name
-        else:
-            changes['shipment_address'] = None
-        if payment_term:
-            changes['payment_term'] = payment_term.id
-            changes['payment_term.rec_name'] = payment_term.rec_name
-        else:
-            changes['payment_term'] = self.default_payment_term()
-        return changes
+                self.payment_term = self.party.customer_payment_term
 
     @fields.depends('currency')
     def on_change_with_currency_digits(self, name=None):
@@ -408,7 +401,7 @@ class Sale(Workflow, ModelSQL, ModelView):
             return self.currency.digits
         return 2
 
-    def get_tax_context(self):
+    def _get_tax_context(self):
         res = {}
         if self.party and self.party.lang:
             res['language'] = self.party.lang.code
@@ -423,92 +416,47 @@ class Sale(Workflow, ModelSQL, ModelView):
 
     @fields.depends('lines', 'currency', 'party')
     def on_change_lines(self):
-        pool = Pool()
-        Tax = pool.get('account.tax')
-        Invoice = pool.get('account.invoice')
-        Configuration = pool.get('account.configuration')
-
-        config = Configuration(1)
-
-        changes = {
-            'untaxed_amount': Decimal('0.0'),
-            'tax_amount': Decimal('0.0'),
-            'total_amount': Decimal('0.0'),
-            }
+        self.untaxed_amount = Decimal('0.0')
+        self.tax_amount = Decimal('0.0')
+        self.total_amount = Decimal('0.0')
 
+        taxes = {}
         if self.lines:
-            context = self.get_tax_context()
-            taxes = {}
-
-            def round_taxes():
-                if self.currency:
-                    for key, value in taxes.iteritems():
-                        taxes[key] = self.currency.round(value)
-
             for line in self.lines:
-                if getattr(line, 'type', 'line') != 'line':
-                    continue
-                changes['untaxed_amount'] += (getattr(line, 'amount', None)
-                    or Decimal(0))
+                self.untaxed_amount += getattr(line, 'amount', None) or 0
+            taxes = self._get_taxes()
+            self.tax_amount = sum(v['amount'] for v in taxes.itervalues())
+        if self.currency:
+            self.untaxed_amount = self.currency.round(self.untaxed_amount)
+            self.tax_amount = self.currency.round(self.tax_amount)
+        self.total_amount = self.untaxed_amount + self.tax_amount
+        if self.currency:
+            self.total_amount = self.currency.round(self.total_amount)
 
-                with Transaction().set_context(context):
-                    tax_list = Tax.compute(getattr(line, 'taxes', []),
-                        getattr(line, 'unit_price', None) or Decimal('0.0'),
-                        getattr(line, 'quantity', None) or 0.0)
-                for tax in tax_list:
-                    key, val = Invoice._compute_tax(tax, 'out_invoice')
-                    if key not in taxes:
-                        taxes[key] = val['amount']
-                    else:
-                        taxes[key] += val['amount']
-                if config.tax_rounding == 'line':
-                    round_taxes()
-            if config.tax_rounding == 'document':
-                round_taxes()
-            changes['tax_amount'] = sum(taxes.itervalues(), Decimal('0.0'))
-        if self.currency:
-            changes['untaxed_amount'] = self.currency.round(
-                changes['untaxed_amount'])
-            changes['tax_amount'] = self.currency.round(changes['tax_amount'])
-        changes['total_amount'] = (changes['untaxed_amount']
-            + changes['tax_amount'])
-        if self.currency:
-            changes['total_amount'] = self.currency.round(
-                changes['total_amount'])
-        return changes
+    @property
+    def taxable_lines(self):
+        taxable_lines = []
+        # In case we're called from an on_change we have to use some sensible
+        # defaults
+        for line in self.lines:
+            if getattr(line, 'type', None) != 'line':
+                continue
+            taxable_lines.append(tuple())
+            for attribute, default_value in (
+                    ('taxes', []),
+                    ('unit_price', Decimal(0)),
+                    ('quantity', 0.0)):
+                value = getattr(line, attribute, None)
+                taxable_lines[-1] += (value
+                    if value is not None else default_value,)
+        return taxable_lines
+
+    @property
+    def tax_type(self):
+        return 'invoice'
 
     def get_tax_amount(self):
-        pool = Pool()
-        Tax = pool.get('account.tax')
-        Invoice = pool.get('account.invoice')
-        Configuration = pool.get('account.configuration')
-
-        config = Configuration(1)
-
-        context = self.get_tax_context()
-        taxes = {}
-
-        def round_taxes():
-            for key, value in taxes.iteritems():
-                taxes[key] = self.currency.round(value)
-
-        for line in self.lines:
-            if line.type != 'line':
-                continue
-            with Transaction().set_context(context):
-                tax_list = Tax.compute(line.taxes, line.unit_price,
-                    line.quantity)
-            for tax in tax_list:
-                key, val = Invoice._compute_tax(tax, 'out_invoice')
-                if key not in taxes:
-                    taxes[key] = val['amount']
-                else:
-                    taxes[key] += val['amount']
-            if config.tax_rounding == 'line':
-                round_taxes()
-        if config.tax_rounding == 'document':
-            round_taxes()
-        return sum(taxes.itervalues(), _ZERO)
+        return sum(v['amount'] for v in self._get_taxes().itervalues())
 
     @classmethod
     def get_amount(cls, sales, names):
@@ -644,6 +592,20 @@ class Sale(Workflow, ModelSQL, ModelView):
                     'shipment_state': state,
                     })
 
+    @classmethod
+    def _get_origin(cls):
+        'Return list of Model names for origin Reference'
+        return ['sale.sale']
+
+    @classmethod
+    def get_origin(cls):
+        Model = Pool().get('ir.model')
+        models = cls._get_origin()
+        models = Model.search([
+                ('model', 'in', models),
+                ])
+        return [(None, '')] + [(m.model, m.name) for m in models]
+
     @classmethod
     def validate(cls, sales):
         super(Sale, cls).validate(sales)
@@ -667,12 +629,20 @@ class Sale(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def search_rec_name(cls, name, clause):
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
         names = clause[2].split(' - ', 1)
-        res = [('reference', clause[1], names[0])]
+        res = [bool_op, ('reference', clause[1], names[0])]
         if len(names) != 1 and names[1]:
             res.append(('party', clause[1], names[1]))
         return res
 
+    @classmethod
+    def view_attributes(cls):
+        return [('//field[@name="comment"]', 'spell', Eval('party_lang'))]
+
     @classmethod
     def copy(cls, sales, default=None):
         if default is None:
@@ -1029,7 +999,7 @@ class SaleLine(ModelSQL, ModelView):
     product_uom_category = fields.Function(
         fields.Many2One('product.uom.category', 'Product Uom Category'),
         'on_change_with_product_uom_category')
-    unit_price = fields.Numeric('Unit Price', digits=(16, 4),
+    unit_price = fields.Numeric('Unit Price', digits=price_digits,
         states={
             'invisible': Eval('type') != 'line',
             'required': Eval('type') == 'line',
@@ -1044,9 +1014,12 @@ class SaleLine(ModelSQL, ModelView):
     description = fields.Text('Description', size=None, required=True)
     note = fields.Text('Note')
     taxes = fields.Many2Many('sale.line-account.tax', 'line', 'tax', 'Taxes',
+        order=[('tax.sequence', 'ASC'), ('tax.id', 'ASC')],
         domain=[('parent', '=', None), ['OR',
                 ('group', '=', None),
                 ('group.kind', 'in', ['sale', 'both'])],
+                ('company', '=',
+                    Eval('_parent_sale', {}).get('company', -1)),
             ],
         states={
             'invisible': Eval('type') != 'line',
@@ -1111,7 +1084,7 @@ class SaleLine(ModelSQL, ModelView):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        return [table.sequence == Null, table.sequence]
 
     @staticmethod
     def default_type():
@@ -1176,6 +1149,7 @@ class SaleLine(ModelSQL, ModelView):
             context['uom'] = self.unit.id
         else:
             context['uom'] = self.product.sale_uom.id
+        context['taxes'] = [t.id for t in self.taxes]
         return context
 
     @fields.depends('product', 'unit', 'quantity', 'description',
@@ -1185,8 +1159,7 @@ class SaleLine(ModelSQL, ModelView):
         Product = Pool().get('product.product')
 
         if not self.product:
-            return {}
-        res = {}
+            return
 
         party = None
         party_context = {}
@@ -1195,71 +1168,70 @@ class SaleLine(ModelSQL, ModelView):
             if party.lang:
                 party_context['language'] = party.lang.code
 
-        category = self.product.sale_uom.category
-        if not self.unit or self.unit not in category.uoms:
-            res['unit'] = self.product.sale_uom.id
-            self.unit = self.product.sale_uom
-            res['unit.rec_name'] = self.product.sale_uom.rec_name
-            res['unit_digits'] = self.product.sale_uom.digits
-
-        with Transaction().set_context(self._get_context_sale_price()):
-            res['unit_price'] = Product.get_sale_price([self.product],
-                    self.quantity or 0)[self.product.id]
-            if res['unit_price']:
-                res['unit_price'] = res['unit_price'].quantize(
-                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
-        res['taxes'] = []
+        # Set taxes before unit_price to have taxes in context of sale price
+        taxes = []
         pattern = self._get_tax_rule_pattern()
         for tax in self.product.customer_taxes_used:
             if party and party.customer_tax_rule:
                 tax_ids = party.customer_tax_rule.apply(tax, pattern)
                 if tax_ids:
-                    res['taxes'].extend(tax_ids)
+                    taxes.extend(tax_ids)
                 continue
-            res['taxes'].append(tax.id)
+            taxes.append(tax.id)
         if party and party.customer_tax_rule:
             tax_ids = party.customer_tax_rule.apply(None, pattern)
             if tax_ids:
-                res['taxes'].extend(tax_ids)
+                taxes.extend(tax_ids)
+        self.taxes = taxes
+
+        category = self.product.sale_uom.category
+        if not self.unit or self.unit not in category.uoms:
+            self.unit = self.product.sale_uom
+            self.unit_digits = self.product.sale_uom.digits
+
+        with Transaction().set_context(self._get_context_sale_price()):
+            self.unit_price = Product.get_sale_price([self.product],
+                    self.quantity or 0)[self.product.id]
+            if self.unit_price:
+                self.unit_price = self.unit_price.quantize(
+                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
 
         if not self.description:
             with Transaction().set_context(party_context):
-                res['description'] = Product(self.product.id).rec_name
+                self.description = Product(self.product.id).rec_name
 
-        self.unit_price = res['unit_price']
         self.type = 'line'
-        res['amount'] = self.on_change_with_amount()
-        return res
+        self.amount = self.on_change_with_amount()
 
     @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
         if self.product:
             return self.product.default_uom_category.id
 
-    @fields.depends('product', 'quantity', 'unit',
+    @fields.depends('product', 'quantity', 'unit', 'taxes',
         '_parent_sale.currency', '_parent_sale.party',
         '_parent_sale.sale_date')
     def on_change_quantity(self):
         Product = Pool().get('product.product')
 
         if not self.product:
-            return {}
-        res = {}
+            return
 
         with Transaction().set_context(
                 self._get_context_sale_price()):
-            res['unit_price'] = Product.get_sale_price([self.product],
+            self.unit_price = Product.get_sale_price([self.product],
                 self.quantity or 0)[self.product.id]
-            if res['unit_price']:
-                res['unit_price'] = res['unit_price'].quantize(
+            if self.unit_price:
+                self.unit_price = self.unit_price.quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
-        return res
 
-    @fields.depends('product', 'quantity', 'unit',
-        '_parent_sale.currency', '_parent_sale.party',
-        '_parent_sale.sale_date')
+    @fields.depends(methods=['quantity'])
     def on_change_unit(self):
-        return self.on_change_quantity()
+        self.on_change_quantity()
+
+    @fields.depends(methods=['quantity'])
+    def on_change_taxes(self):
+        self.on_change_quantity()
 
     @fields.depends('type', 'quantity', 'unit_price', 'unit',
         '_parent_sale.currency')
@@ -1365,8 +1337,15 @@ class SaleLine(ModelSQL, ModelView):
             if old_invoice_line.type != 'line':
                 continue
             if old_invoice_line.id not in skip_ids:
+                if old_invoice_line.invoice:
+                    old_invoice_type = old_invoice_line.invoice.type
+                else:
+                    old_invoice_type = old_invoice_line.invoice_type
+
+                sign = 1 if old_invoice_type == invoice_type else -1
+
                 quantity -= Uom.compute_qty(old_invoice_line.unit,
-                    old_invoice_line.quantity, self.unit)
+                    sign * old_invoice_line.quantity, self.unit)
 
         rounding = self.unit.rounding if self.unit else 0.01
         invoice_line.quantity = Uom.round(quantity, rounding)
@@ -1396,6 +1375,11 @@ class SaleLine(ModelSQL, ModelView):
                         })
         return [invoice_line]
 
+    @classmethod
+    def view_attributes(cls):
+        return [('//field[@name="note"]|//field[@name="description"]',
+                'spell', Eval('_parent_sale', {}).get('party_lang'))]
+
     @classmethod
     def copy(cls, lines, default=None):
         if default is None:
@@ -1566,7 +1550,7 @@ class HandleShipmentException(Wizard):
     ask = StateView('sale.handle.shipment.exception.ask',
         'sale.handle_shipment_exception_ask_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Ok', 'handle', 'tryton-ok', default=True),
+            Button('OK', 'handle', 'tryton-ok', default=True),
             ])
     handle = StateTransition()
 
@@ -1622,7 +1606,7 @@ class HandleInvoiceExceptionAsk(ModelView):
         domain=[('id', 'in', Eval('domain_invoices'))],
         depends=['domain_invoices'],
         help='The selected invoices will be recreated. '
-            'The other ones will be ignored.')
+        'The other ones will be ignored.')
     domain_invoices = fields.Many2Many(
         'account.invoice', None, None, 'Domain Invoices')
 
@@ -1634,7 +1618,7 @@ class HandleInvoiceException(Wizard):
     ask = StateView('sale.handle.invoice.exception.ask',
         'sale.handle_invoice_exception_ask_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Ok', 'handle', 'tryton-ok', default=True),
+            Button('OK', 'handle', 'tryton-ok', default=True),
             ])
     handle = StateTransition()
 
@@ -1698,11 +1682,14 @@ class ReturnSale(Wizard):
 
         sales = Sale.browse(Transaction().context['active_ids'])
         return_sales = Sale.copy(sales)
-        for sale in return_sales:
-            for line in sale.lines:
+        for return_sale, sale in zip(return_sales, sales):
+            return_sale.origin = sale
+            for line in return_sale.lines:
                 if line.type == 'line':
                     line.quantity *= -1
-                    line.save()
+            return_sale.lines = return_sale.lines  # Force saving
+        Sale.save(return_sales)
+
         data = {'res_id': [s.id for s in return_sales]}
         if len(return_sales) == 1:
             action['views'].reverse()
